/*
  RabidVWM LiteStep Module - version 1.0
	Pixel-scale Bitmap Manipulation

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "RabidVWM.h"
#include "bitfiddle.h"

// bitmap tricks

COLORREF *PixelBlock::Pixel(int x,int y)
{
	ASSERTV(hBmp!=NULL);
	if ( x<0
		|| x>Bmp.bmWidth
		|| y<0
		|| y>Bmp.bmHeight)
		return NULL;

	return 
		(GetBits()
		+ (Bmp.bmHeight-1-y)*Bmp.bmWidthBytes/sizeof(COLORREF)
		+ x);
}

PixelBlock::PixelBlock(PixelBlock &src)
{
	hBmp = src.Release();
	CopyMemory(&Bmp,&src.Bmp,sizeof(Bmp));
}

// create an uninitialized block of pixels
PixelBlock::PixelBlock(DWORD width,DWORD height)
{
	BITMAPINFOHEADER bmInfo;
	void *bits;

	// set everything to 0
	ZeroMemory(&bmInfo, sizeof(BITMAPINFOHEADER) );

	// Create a 32 bits depth bitmap
	bmInfo.biSize = sizeof(BITMAPINFOHEADER);
	bmInfo.biWidth = width;
	bmInfo.biHeight = height;
	bmInfo.biPlanes = 1;
	bmInfo.biBitCount = 32;
	bmInfo.biCompression = BI_RGB;

	// create the bitmap
	hBmp =
	CreateDIBSection(
		NULL,
		(BITMAPINFO*)&bmInfo,
		DIB_RGB_COLORS,
		&bits,
		NULL,
		0);
	StatGDIAdd(Bitmap);

	// get the bits from hbm32
	GetObject(hBmp, sizeof(Bmp), &Bmp );
}

// create a block of pixels from a bitmap handle
// the bitmap passed to this is destroyed
// (absorbed, if you like)
PixelBlock::PixelBlock(HBITMAP hbm)
{
	// this code is adapted from grdtray

	HDC srcDC,dstDC;		// device contexts for copying
	BITMAP bm;				// Get bitmap size
	BITMAPINFOHEADER bmInfo;// specify requested bmp params
	VOID *pbits32;			// never really use it

	// get bitmap size
	GetObject(hbm, sizeof(bm), &bm ); 

	// is it already 32bpp?
	if (bm.bmBitsPixel==32
		&& bm.bmBits!=NULL) {

		// just assign from hbm
		CopyMemory(&Bmp,&bm,sizeof(bm));
		hBmp = hbm;
		return;
	}

	srcDC = CreateCompatibleDC(NULL);
	StatGDIAdd(DC);
	HBITMAP hbmOldSrc;
	// select the original bitmap into its dc
	hbmOldSrc = (HBITMAP)SelectObject(srcDC, hbm);

	// set everything to 0
	ZeroMemory(&bmInfo, sizeof(BITMAPINFOHEADER) );

	// Create a 32 bits depth bitmap
	bmInfo.biSize = sizeof(BITMAPINFOHEADER);
	bmInfo.biWidth = bm.bmWidth;
	bmInfo.biHeight = bm.bmHeight;
	bmInfo.biPlanes = 1;
	bmInfo.biBitCount = 32;
	bmInfo.biCompression = BI_RGB;

	hBmp = CreateDIBSection(
		srcDC,
		(BITMAPINFO*) &bmInfo,
		DIB_RGB_COLORS,
		&pbits32,
		NULL,
		0);
	StatGDIAdd(Bitmap);

	if (hBmp) {
		// create dcs for copying,
		// compatible to the screen
		dstDC = CreateCompatibleDC(NULL);
		StatGDIAdd(DC);

		// select the new bitmap into its dc
		HBITMAP hbmOldDst
			=(HBITMAP)SelectObject(dstDC, hBmp);

		// get the bits from hbm32
		GetObject(hBmp, sizeof(Bmp), &Bmp );

		// Get how many bytes per row we have for
		// the bitmap bits (rounded up to 32 bits)
		while (Bmp.bmWidthBytes % 4 )
			Bmp.bmWidthBytes++;

		// copy the original bitmaps contents
		// into the new one
		BitBlt(dstDC, 0, 0, bm.bmWidth, bm.bmHeight,
			srcDC, 0, 0, SRCCOPY);

		// Select the bitmaps out of their dc's
		SelectObject(dstDC,hbmOldDst);

		// destroy temporary dc's
		DeleteDC(dstDC);
		StatGDIDel(DC);
	}
	SelectObject(srcDC,hbmOldSrc);
	DeleteDC(srcDC); StatGDIDel(DC);
	DeleteObject(hbm); StatGDIDel(Bitmap);
}

PixelBlock::~PixelBlock()
{
	if (hBmp!=NULL)
	{
		DeleteObject(Release());
		StatGDIDel(Bitmap);
	}
}

HBITMAP PixelBlock::Release(void)
{
	HBITMAP hbm = hBmp;
	hBmp = NULL;
	return hbm;
}



void PixelBlockIterator::operator ++(void)
{
	ASSERTV(!Done());
	ptr++;
	if (++x>=(DWORD)area.right) {
		// go to beginning of row, then up(down) one
		ptr-=x+src.GetWidthBytes()/sizeof(COLORREF);
		// now go to start of section
		ptr+=area.left;
		x=area.left;
		if (++y>=(DWORD)area.bottom) {
			y=area.top;
			ptr = NULL;
		}
	}
}

// get pixel by offset
COLORREF PixelBlockIterator::Get(int x,int y)
{
	ASSERTV(!Done());
	return src.GetPixel(x+this->x,y+this->y);
}

