/*
  RabidVWM LiteStep Module - version 1.0
	Virtual Desktop Functions

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
		When it comes to switching desktops, there are two methods:
		you can hide windows not on the current desktop, or you can
		move them.
		
		Hiding will remove windows from the taskbar, and could
		potentially lose them if LiteStep crashes, but there are
		some windows that moving doesn't work on.  (because they
		override SetWindowPos to stay on screen, often the primary
		monitor too)

		Moving will never lose any windows, but as I said above, it
		doesn't always work.  Also, some weirdos seem to like having
		things on different desktops not appear in the taskbar or
		alt-tab list.


		Hiding actually has another disadvantage, which appears on
		recycles.  The VWM has to make sure it can pick up all the
		windows again, so all the windows must be shown before the
		module is unloaded.

		(Note that the actual hide/move is decided in MoveWindowVWM.)

*/

#include "RabidVWM.h"

// Input:
//  desk = desk number
// Output:
//  off = x and y coords of top-left corner of desk
void GetDeskOff(int desk,POINT *off)
{
	ASSERTV(desk>=0);
	ASSERTV(desk<GlobalData.nDesks);
	ASSERT(WRITEABLE(off,sizeof(*off)));
  ASSERTV(GlobalData.nDesksX!=0);

	// same as GetDeskDif(CurrentDesk,desk,off) and offsetting by DeskOffset
	off->x = ((GlobalData.CurrentDesk%GlobalData.nDesksX)-(desk%GlobalData.nDesksX))
		* GlobalData.DeskSize.cx + GlobalData.DeskOffset.x;
	off->y = ((GlobalData.CurrentDesk/GlobalData.nDesksX)-(desk/GlobalData.nDesksX))
		* GlobalData.DeskSize.cy + GlobalData.DeskOffset.y;
}

// Input:
//	olddesk, newdesk
// Output:
//	dif = difference in coords, newdesk-olddesk
void GetDeskDif(int olddesk,int newdesk,SIZE *dif)
{
	ASSERT(WRITEABLE(dif,sizeof(*dif)));

	GetDeskDifXY(
		(olddesk%GlobalData.nDesksX),
		(olddesk/GlobalData.nDesksX),
		(newdesk%GlobalData.nDesksX),
		(newdesk/GlobalData.nDesksX),
		dif);
}

void GetDeskDifXY(int oldx,int oldy,int newx,int newy,SIZE *dif)
{
	ASSERT(WRITEABLE(dif,sizeof(*dif)));

	dif->cx = (oldx-newx) * GlobalData.DeskSize.cx;
	dif->cy = (oldy-newy) * GlobalData.DeskSize.cy;
}

void SetDeskFocus(void)
{
	HWND hWnd;
	s_WinData *pData;
	int i;
  bool bNoRetry = false;

	/* kill any active focus switch timers                */ 
	KillTimer(GlobalData.hMainWnd,GlobalData.FocusTimerID);

	/* don't switch focus at all                          */ 
	if (!GlobalData.bFocusTop && !GlobalData.bFocusLast) return;

	/* no windows to switch to                            */ 
	if (GlobalData.nKnownWindows==0) return;

	/* is the foreground window already on this desk?     */ 
	hWnd = GetForegroundWindow();
	pData = GetWindowPtr(hWnd);
	WTRACE(TRACE_STATUS,pData,"Had focus");
	if (!(hWnd==pData->hWnd && pData->HomeDesk==GlobalData.CurrentDesk))
	{
		/* switch to foreground window on this desk         */ 
		hWnd = NULL;
		/* first try last window on this desk               */ 
		if (hWnd==NULL && GlobalData.bFocusLast)
		{
			hWnd = GlobalData.Desks[GlobalData.CurrentDesk].hLastActiveWnd;
			pData = GetWindowPtr(hWnd);
			/* if it still exists, and is still on this desk  */ 
			if (pData->hWnd!=hWnd || pData->HomeDesk!=GlobalData.CurrentDesk)
				/* otherwise forget it                          */ 
				hWnd = NULL;
		}
		/* then topmost window on this desk                 */ 
		if (hWnd==NULL && GlobalData.bFocusTop)
		{
			/* search from back to front                      */ 
			i = GlobalData.FirstWindow;
			while (i>=0) {
				pData = &GlobalData.KnownWindows[i];
				if (pData->HomeDesk==GlobalData.CurrentDesk
				  /* don't switch to minimized windows this way   */ 
				  && !IsWndMinimized(pData)
				  /* only switch to topmost if nothing else there */ 
          && (!IsTopMost(pData->hWnd) || hWnd==NULL))
				{
					hWnd = pData->hWnd;
				}
				i = pData->PaintNext;
			}
		}
		/* switch to vwm if no windows available            */ 
		if (hWnd==NULL) 
    {
      TRACE(TRACE_STATUS,"No window found to switch to.");
      bNoRetry = true;
      hWnd = GlobalData.hMainWnd;
		  if (GlobalData.bStayChildWindow)
        hWnd = GlobalData.hParentWnd;
    }
    /* should never happen, but if all else fails switch */ 
    /* to main litestep window (should use desktop?)     */ 
		if (hWnd==NULL) hWnd = GlobalData.hLiteStepWnd;

		hWnd = GetLastActivePopup(hWnd);

#if !defined(NDEBUG) && defined(TRACE_FILE)
		{
			char buf[256];
			sprintf(buf,"Switching to %08X",hWnd);
      if (hWnd==GlobalData.hMainWnd) strcat(buf," (VWM)");
      else if (hWnd==GlobalData.hParentWnd) strcat(buf," (parent)");
      else if (hWnd==GlobalData.hLiteStepWnd) strcat(buf," (litestep)");
      else strcat(buf," (no idea)");
			TRACE(TRACE_STATUS,buf);
			*buf='$'; GetWindowText(hWnd,buf+1,255);
			if (!buf[1]) { *buf='@'; GetClassName(hWnd,buf+1,255); }
			TRACE(TRACE_STATUS,buf);
		}
#endif

		if (!IsWindow(hWnd)) {
			TRACE(TRACE_WARNING,"Window that would gain focus doesn't exist.");
			return;
		}

		/* AFAIK, fAltTab (the second param) should be FALSE, but for some reason    */ 
		/* Windows sometimes decides that that means the window should be forced     */ 
		/* beneath the desktop window.                                               */ 

		/* switch for winnt/9x                                                       */ 
		if (SwitchToThisWindow) SwitchToThisWindow(hWnd,TRUE);


		/* This next mess shouldn't be necessary, but sometimes SwitchToThisWindow   */ 
		/* doesn't seem to work right away, but after a delay it will.  Might have   */ 
		/* something to do with the window being outside of the visible area and not */ 
		/* processing its message loop fast enough, but I don't think that should be */ 
		/* a problem unless it's forced to move async.                               */ 
		
		/* switch for win95                                                          */ 
		if (!bNoRetry && !SetForegroundWindow(hWnd))
		{
      if (SwitchToThisWindow)
      {
			  /* set focus to vwm to keep from switching desks                           */ 
			  if (GlobalData.hParentWnd) /* mustn't switch to child window               */ 
				  /* switching to child window confused lsbox, and probably anything else  */ 
				  /* or maybe it confused windows... bad things, trust me.                 */ 
				  SwitchToThisWindow(GlobalData.hParentWnd,TRUE);
			  else
				  SwitchToThisWindow(GlobalData.hMainWnd,FALSE);
      }

			/* set a timer to try again                                                */ 
			TRACE(TRACE_WARNING,"Will try again in 500 ms");
			GlobalData.FocusTimerID = SetTimer(GlobalData.hMainWnd,TIMER_FOCUS,500,NULL);
		}

	}
	else
	{
		TRACE(TRACE_STATUS,"Foreground window is already on this desk.");
	}
}

// move all windows over to switch to the specified desktop and set flag for a repaint
bool SwitchDesk(int newdesk)
{
	static volatile bool bInSwitchDesk=false;  /* don't switch desks when i'm halfway done */ 
	/* this is done to prevent losing track of desks with FocusXXX and SwitchOnAnyFocus    */ 
	/* (SwitchOnAnyFocus would trigger another SwitchDesk before the first was done)       */ 
	/* should probably use real locks or whatever, but hey this works for now              */ 

	SIZE dif;
	POINT cursorpos;

	ASSERTV(newdesk>=0);
	ASSERTV(newdesk<GlobalData.nDesks);
	ASSERTV(GlobalData.CurrentDesk>=0);
	ASSERTV(GlobalData.CurrentDesk<GlobalData.nDesks);
	ASSERTV(GlobalData.Desks!=NULL);

	if (bInSwitchDesk) return false;
	bInSwitchDesk = true;

#if !defined(NDEBUG) && defined(TRACE_FILE)
	{
		char buffer[64];
		sprintf(buffer,"SwitchDesk(%d)",newdesk);
		TRACE(TRACE_STATUS,buffer);
	}
#endif

  if (GlobalData.nSetupPass==0)
	  AtVarExec(GlobalData.OnSwitchFrom,false,false);

	/* if a window is being dragged, leave it here */ 
	/* also, don't switch focus when we get to the */ 
	/* new desk                                    */ 
	if (MouseButtonDown())
	{
		GetCursorPos(&cursorpos);
		GlobalData.hWndDragging = WindowFromPoint(cursorpos);
		if (GlobalData.hFocusWnd!=GetForegroundWindow())
			GlobalData.hWndDragging = NULL;
	}

	/* save the current foreground window          */ 
	GlobalData.Desks[GlobalData.CurrentDesk].hLastActiveWnd = GlobalData.hFocusWnd;

	GetDeskDif(GlobalData.CurrentDesk,newdesk,&dif);

	if (MoveAllWindows(&dif,newdesk)) {
		GlobalData.CurrentDesk = newdesk;

		if (GlobalData.hWndDragging==NULL
      || GlobalData.hWndDragging==GlobalData.hMainWnd)
			SetDeskFocus();

		GlobalData.hWndDragging = NULL;

		/* these are cleared at the end of repainting                   */ 
		/* make sure the desk changes even if there're no windows       */ 
    TRACE(TRACE_STATUS,"bChangeState due to SwitchDesk");
		GlobalData.bChangeState = true;
		/* slow updating windows need to wait a bit to get the snapshot */ 
		GlobalData.bDelaySnapshot = true;

    if (GlobalData.nSetupPass==0)
		  AtVarExec(GlobalData.OnSwitchTo,false,false);

		/* make sure things get updated                                 */ 
		UpdateWindowList();

		bInSwitchDesk = false;
		return true;
	} else {
		GlobalData.hWndDragging = NULL;

		/* whether or not to do this is debatable, but I think it's     */ 
		/* better to, that way you never get two switch froms in a row  */ 
    if (GlobalData.nSetupPass==0)
		  AtVarExec(GlobalData.OnSwitchTo,false,false);

		TRACE(TRACE_WARNING,"Failed to switch desktops.");
		bInSwitchDesk = false;
		return false;
	}
}

// return 0-based desk number containing point x,y
// or -1 if x,y is not inside any desk
int DeskFromPoint(int x,int y)
{
	POINT zero;

  ASSERTV(GlobalData.DeskSize.cx!=0);
  ASSERTV(GlobalData.DeskSize.cy!=0);

	/* get absolute desk coords                              */ 
	GetDeskOff(0,&zero);
	zero.x += x;
	zero.y += y;

	/* make sure out of vwm coords return an invalid desk no */ 
	if (zero.x<0 || zero.y<0) return DESK_NONE;

	zero.x /= GlobalData.DeskSize.cx;
	zero.y /= GlobalData.DeskSize.cy;

	if (zero.x>=GlobalData.nDesksX ||zero.y>=GlobalData.nDesksY)
    return DESK_NONE;

	/* convert x,y to desk number                            */ 
	return zero.x + zero.y*GlobalData.nDesksX;
}

// returns x,y desk coords from point x,y
// does not check against valid desk coords.
void DeskFromPointXY(int x,int y,POINT *desk)
{
	POINT fix;

	ASSERT(WRITEABLE(desk,sizeof(*desk)));
  ASSERTV(GlobalData.DeskSize.cx!=0);
  ASSERTV(GlobalData.DeskSize.cy!=0);

	/* get absolute desk coords                              */ 
	GetDeskOff(0,desk);
	desk->x += x;
	desk->y += y;

	/* If it's lost off the top or left edge, need to adjust.     */ 
	/* What I really need is to round towards -inf rather than 0. */ 
	if (desk->x<0) fix.x=1; else fix.x=0;
	if (desk->y<0) fix.y=1; else fix.y=0;

	desk->x /= GlobalData.DeskSize.cx;
	desk->y /= GlobalData.DeskSize.cy;

	desk->x -= fix.x;
	desk->y -= fix.y;
}

// returns 0-based desk number containing x,y on the VWM
// or -1 if x,y is not inside a valid desk on the VWM
int DeskFromVWMPoint(int x,int y)
{
	/* offset by start of vwm          */ 
	x -= GlobalData.rDeskBorder.left;
	y -= GlobalData.rDeskBorder.top;
	/* convert vwm coord to desk coord */ 
	x = x/GlobalData.VWMDeskSize.cx;
	y = y/GlobalData.VWMDeskSize.cy;

	if (x<0 || y<0) return DESK_NONE;
	if (x>=GlobalData.nDesksX) return DESK_NONE;
	if (y>=GlobalData.nDesksY) return DESK_NONE;

	/* and then to desk #              */ 
	return y*GlobalData.nDesksX+x;
}

// get the VWM rect for a desk
void GetDeskVWMRect(int desk, RECT *r)
{
	ASSERT(WRITEABLE(r,sizeof(*r)));
  ASSERTV(GlobalData.nDesksX!=0);

	r->left = GlobalData.CurrentDesk%GlobalData.nDesksX;
	r->left *= GlobalData.VWMDeskSize.cx;
	r->left += GlobalData.rDeskBorder.left;

	r->top = GlobalData.CurrentDesk/GlobalData.nDesksX;
	r->top *= GlobalData.VWMDeskSize.cy;
	r->top += GlobalData.rDeskBorder.top;

	r->right = r->left + GlobalData.VWMDeskSize.cx;
	r->bottom = r->top + GlobalData.VWMDeskSize.cy;
}

#if(0) // never used
bool OutsideAnyDesk(const POINT *pt)
{
	POINT zero;

	ASSERT(READABLE(pt,sizeof(*pt)));

	/* get absolute desk coords */ 
	GetDeskOff(0,&zero);
	zero.x += pt->x;
	zero.y += pt->y;

	return
		((zero.x<0)
		||(zero.y<0)
		||(zero.x>=(GlobalData.nDesksX*GlobalData.DeskSize.cx))
		||(zero.y>=(GlobalData.nDesksY*GlobalData.DeskSize.cy)));
}

bool IsOnScreen(const RECT *area)
{
	RECT screen;
	POINT tl,tr,bl,br;

	ASSERT(READABLE(area,sizeof(*area)));

	screen.left = GlobalData.DeskOffset.x;
	screen.top = GlobalData.DeskOffset.y;
	screen.right = screen.left + GlobalData.DeskSize.cx;
	screen.bottom = screen.top + GlobalData.DeskSize.cy;
	
	tl.x = bl.x = area->left;
	tl.y = tr.y = area->top;
	br.x = tr.x = area->right;
	br.y = bl.y = area->bottom;
	/* if any corner of area is on screen */ 
	if (PtInRect(&screen,tl)
		||PtInRect(&screen,tr)
		||PtInRect(&screen,bl)
		||PtInRect(&screen,br))
		return true;

	/* Under normal circumstances, that should be enough, but */ 
	/* sometimes windows are accidentaly positioned wrong     */ 

	/* if any edge of area is on screen                       */ 

	/* horizontal edge                                        */ 
	if ((area->left<screen.left && area->right>=screen.right)
		&&((area->top>screen.top && area->top<=screen.bottom)
			||(area->bottom>screen.top && area->bottom<=screen.bottom)))
		return true;

	/* vertical edge                                          */ 
	if ((area->top<screen.top && area->bottom>=screen.bottom)
		&&((area->left>screen.left && area->left<=screen.right)
			||(area->right>screen.left && area->right<=screen.right)))
		return true;

	/* if screen is completely enclosed in area               */ 
	/* any corner of the screen should work for this          */ 
	if (PtInRect(area,GlobalData.DeskOffset))
		return true;

	return false;
}
#endif

void ScreenToVWM(RECT *dest,const RECT *source)
{
	POINT zero;

	ASSERT(READABLE(dest,sizeof(*dest)));
	ASSERT(WRITEABLE(dest,sizeof(*dest)));
	ASSERT(READABLE(source,sizeof(*source)));
  ASSERTV(GlobalData.DeskSize.cx!=0);
  ASSERTV(GlobalData.DeskSize.cy!=0);

	/* get absolute desk coords                                           */ 
	GetDeskOff(0,&zero);

	/* convert width rather than point to make it independant of position */ 
	dest->right = source->right-source->left;
	dest->bottom = source->bottom-source->top;

	dest->left = source->left + zero.x;
	dest->top = source->top + zero.y;

	/* scale to vwm                                                       */ 
	dest->left = MulDiv(dest->left,GlobalData.VWMDeskSize.cx,GlobalData.DeskSize.cx)
		+ GlobalData.rDeskBorder.left;
	dest->top = MulDiv(dest->top,GlobalData.VWMDeskSize.cy,GlobalData.DeskSize.cy)
		+ GlobalData.rDeskBorder.top + 1;
	dest->right = dest->left
		+ MulDiv(dest->right,GlobalData.VWMDeskSize.cx,GlobalData.DeskSize.cx);
	dest->bottom = dest->top
		+ MulDiv(dest->bottom,GlobalData.VWMDeskSize.cy,GlobalData.DeskSize.cy);
}

void VWMToScreen(int x,int y,POINT *pt)
{
	POINT zero;

	ASSERT(WRITEABLE(pt,sizeof(*pt)));

	/* get absolute desk coords */ 
	GetDeskOff(0,&zero);

	pt->x = (x-GlobalData.rDeskBorder.left)*GlobalData.Scale.cx-zero.x;
	pt->y = (y-GlobalData.rDeskBorder.top)*GlobalData.Scale.cy-zero.y;
}

// get an absolute desk number from relative desk coordinates
// x and y are in desktops, not points
// this will wrap if set, or return an invalid desktop no otherwise
int GetDeskFromDelta(int x,int y)
{
	int dx,dy;

  ASSERTV(GlobalData.nDesksX!=0);

	/* get desk */ 
	dx = GlobalData.CurrentDesk % GlobalData.nDesksX;
	dy = GlobalData.CurrentDesk - dx + GlobalData.nDesksX * y;
	dx += x;

	if (!GlobalData.bNoWrap)
	{
		/* wrap */ 
		while (dx<0) dx += GlobalData.nDesksX;
		while (dy<0) dy += GlobalData.nDesks;
		if (dx>=GlobalData.nDesksX) dx %= GlobalData.nDesksX;
		if (dy>=GlobalData.nDesks) dy %= GlobalData.nDesks;
	}
	else
	{
		/* clip */ 
		if (dx<0) dx = 0;
		if (dy<0) dy = 0;
		if (dx>=GlobalData.nDesksX) dx = GlobalData.nDesksX-1;
		if (dy>=GlobalData.nDesksY) dy = GlobalData.nDesksY-1;
	}

	/* destroy the planet */ 
	return dx+dy;
}

// set the home desk for a window and keep the windows per desk count valid
// (windows/desk is primarily used for ^@de^)
void SetHomeDesk(s_WinData *pData,int home)
{
	ASSERTR(READABLE(pData,sizeof(*pData)),;);
	ASSERTR(WRITEABLE(pData,sizeof(*pData)),;);
	ASSERTV(home<GlobalData.nDesks);
  ASSERT(WRITEABLE(GlobalData.Desks,GlobalData.nDesks*sizeof(*GlobalData.Desks)));

	if (pData->HomeDesk>=0)
		GlobalData.Desks[pData->HomeDesk].nWnds--;
	pData->HomeDesk = home;
	if (pData->HomeDesk>=0)
		GlobalData.Desks[pData->HomeDesk].nWnds++;
}

// guess number and position of desks based on current windows
// intended as a crash recovery, to help prevent large numbers
// of lost windows.
void GuessDesksFromWindows(POINT *dmin,POINT *dmax)
{
	int i;
	s_WinData *pData;
	POINT desk;

	ASSERT(WRITEABLE(dmin,sizeof(dmin)));
	ASSERT(WRITEABLE(dmax,sizeof(dmax)));
  ASSERT(READABLE(GlobalData.KnownWindows,GlobalData.nKnownWindows*sizeof(*GlobalData.KnownWindows)));

	TRACE(TRACE_STATUS,"GuessDesksFromWindows");

	dmin->x = dmax->x =
	dmin->y = dmax->y = 0;
	pData = GlobalData.KnownWindows; i = GlobalData.nKnownWindows;
	while (--i>=0) {
		DeskFromPointXY(
			(pData->Area.right+pData->Area.left)/2,
			(pData->Area.top+pData->Area.bottom)/2,
			&desk);
		if (desk.x < dmin->x) dmin->x = desk.x;
		if (desk.y < dmin->y) dmin->y = desk.y;
		if (desk.x > dmax->x) dmax->x = desk.x;
		if (desk.y > dmax->y) dmax->y = desk.y;
		WTRACE(TRACE_STATUS,pData,"checking");
		pData++;
	}

#if !defined(NDEBUG) && defined(TRACE_FILE)
	{
		char buf[256];
		sprintf(buf,"Hmm, looks like %d x %d desks, starting at %d %d.",
			1+dmax->x-dmin->x,1+dmax->y-dmin->y,
			1-dmin->x,1-dmin->y);
		TRACE(TRACE_STATUS,buf);
	}
#endif
}

// sets desk size and desk to screen scales
void SetScreenToDeskConstants(void)
{
	ASSERTV(GlobalData.nDesksX!=0);
	ASSERTV(GlobalData.nDesksY!=0);
  /* ASSERT(GlobalData.VWMDeskSize.cx!=0); ok. */ 
  /* ASSERT(GlobalData.VWMDeskSize.cy!=0); ok. */ 

	/* screen constants */ 
	TRACE(TRACE_STATUS,"Setting up mini-screen constants.");
	GlobalData.VWMDeskSize.cx =
		(GlobalData.VWMSize.cx-GlobalData.rDeskBorder.left-GlobalData.rDeskBorder.right)/
			GlobalData.nDesksX;
	GlobalData.VWMDeskSize.cy =
		(GlobalData.VWMSize.cy-GlobalData.rDeskBorder.top-GlobalData.rDeskBorder.bottom)/
			GlobalData.nDesksY;

	/* the desk size could be 0, but only if you're crazy     */ 
	/* the vwm won't work right in that case, but don't crash */ 
	if (GlobalData.VWMDeskSize.cx==0) GlobalData.Scale.cx = 0;
	else GlobalData.Scale.cx = GlobalData.DeskSize.cx/GlobalData.VWMDeskSize.cx;
	if (GlobalData.VWMDeskSize.cy==0) GlobalData.Scale.cy = 0;
	else GlobalData.Scale.cy = GlobalData.DeskSize.cy/GlobalData.VWMDeskSize.cy;

	/* should be done at most once, not each time unless it gets reset
	// adjust outside border to hide rounding errors
	GlobalData.rDeskBorder.left +=
		(GlobalData.VWMSize.cx-GlobalData.rDeskBorder.left-GlobalData.rDeskBorder.right
			-GlobalData.VWMDeskSize.cx*GlobalData.nDesksX)/2;
	GlobalData.rDeskBorder.top +=
		(GlobalData.VWMSize.cy-GlobalData.rDeskBorder.top-GlobalData.rDeskBorder.bottom
			-GlobalData.VWMDeskSize.cy*GlobalData.nDesksY)/2;
	*/
}

// set or change the number of desktops
// this will not resize the main window
void SetDeskCount(int x,int y)
{
	int i;
	s_WinData *pData;
	s_DeskData *Desks;

	ASSERTV(x!=0);
	ASSERTV(y!=0);

	/* calculate number of desks                   */ 
	i = x * y;

	/* allocate desk data structures               */ 
	Desks = new s_DeskData[i];
	/* fail if we couldn't allocate the desk array */ 
	/* this will leave the old settings, if any    */ 
	if (Desks==NULL) return;
	ZeroMemory(Desks,i*sizeof(s_DeskData));

	if (GlobalData.Desks) {
		delete[] GlobalData.Desks;
		DCLEAR(GlobalData.Desks);
	}

	GlobalData.nDesksX = x;
	GlobalData.nDesksY = y;
	GlobalData.nDesks = i;
	GlobalData.Desks = Desks;

	SetScreenToDeskConstants();

	/* count windows/desk                          */ 
	pData = GlobalData.KnownWindows;
	i = GlobalData.nKnownWindows;
	while (--i>=0) {
		if (pData->HomeDesk>=0 && pData->HomeDesk<GlobalData.nDesks) {
			GlobalData.Desks[pData->HomeDesk].nWnds++;
		}
		/* make sure it's different so it gets updated */ 
		pData->Area.right++;
		UpdateRect(pData);
		/* I always forget this bit...                 */ 
		pData++;
	}

	/* i don't usually check for this, so make sure we get repainted */ 
  TRACE(TRACE_STATUS,"bChangeState due to SetDeskCount");
	GlobalData.bChangeState = true;
}

// change the size of the vwm window
// this doesn't change the number of desktops
void Resize(int width, int height)
{
	HBITMAP bm;
	HDC dc;
	int i;
	s_WinData *pData;

	ASSERTV(GlobalData.hMainWnd!=NULL);

	TRACE(TRACE_STATUS,"Creating back buffer.");
	dc = GetDC(GlobalData.hMainWnd);
	CTRACE(TRACE_WARNING,!dc,"Couldn't get MainWnd DC.");
	bm = CreateCompatibleBitmap(dc,width,height);
	StatGDIAdd(Bitmap);
	ReleaseDC(GlobalData.hMainWnd,dc);

	if (!bm) { TRACE(TRACE_WARNING,"failed."); return; }

	if (GlobalData.bmBuffer)
	{
		DeleteObject(GlobalData.bmBuffer);
		StatGDIDel(Bitmap);
	}
	GlobalData.bmBuffer = bm;
	GlobalData.VWMSize.cx = width;
	GlobalData.VWMSize.cy = height;

	SetScreenToDeskConstants();

	SetWindowPos(GlobalData.hMainWnd,NULL,0,0,width,height,
		SWP_NOMOVE|SWP_NOACTIVATE|SWP_NOZORDER);

	/* update miniwindow rectangles */ 
	GlobalData.bScaleChanged=true;
	pData = GlobalData.KnownWindows;
	i = GlobalData.nKnownWindows;
	while (--i>=0) {
		UpdateRect(pData);
		pData++;
	}

	UpdateDesktopBits();
	Repaint();
}

