/*
  RabidVWM LiteStep Module - version 1.0
	Read and Release Configuration

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "RabidVWM.h"
#include "multimonitor.h"
#include <stdlib.h>

// paint-colors.cpp
void PaintVWM_Colors(HDC dc);
void PaintSelect_Colors(HDC,s_DeskData *,RECT *);
// paint-bitmap.cpp
void PaintVWM_Bitmap(HDC dc);
void PaintSelect_Bitmap(HDC,s_DeskData *,RECT *);
// paint-snapshot.cpp
void PaintVWM_SnapShot(HDC dc);
void PaintSelect_SnapShot(HDC,s_DeskData *,RECT *);


s_ParseData *g_pParseData = NULL;

static void GetRCPointS(
	const char *key, int keylen,
	const char *x, int xmin, int xmax,
	const char *y, int ymin, int ymax,
	POINT *pt)
{
	int args;

	ASSERTV(key!=NULL);
	ASSERTV(pt!=NULL);
	ASSERTV(g_pParseData!=NULL);

	lstrcpy(g_pParseData->keybuf,key);
	if (GetRCLine(g_pParseData->keybuf,g_pParseData->line,MAX_LINE_LENGTH,""))
	{
		args = CommandTokenize(g_pParseData->line,g_pParseData->tokens,2,NULL);
		if (args>0) pt->x = dParseCoordinate(g_pParseData->one,pt->x-xmin,xmax-xmin)+xmin;
		if (args>1) pt->y = dParseCoordinate(g_pParseData->two,pt->y-ymin,ymax-ymin)+ymin;
	}
	else
	{
		/* try reading separate x and y                  */ 
		lstrcpy(g_pParseData->keybuf+keylen,x);
		if (GetRCLine(g_pParseData->keybuf,g_pParseData->line,MAX_LINE_LENGTH,""))
			pt->x = dParseCoordinate(g_pParseData->line,pt->x-xmin,xmax-xmin)+xmin;

		lstrcpy(g_pParseData->keybuf+keylen,y);
		if (GetRCLine(g_pParseData->keybuf,g_pParseData->line,MAX_LINE_LENGTH,""))
			pt->y = dParseCoordinate(g_pParseData->line,pt->y-ymin,ymax-ymin)+ymin;
	}
}

/*
	Reads x and y values using ParseXCoord and ParseYCoord
	from key or key+"X" and key+"Y"
*/
static void GetRCPoint(const char *key,int keylen,POINT *pt,s_GlobalData *pGData)
{
	GetRCPointS(key,keylen,
		"X",pGData->rcScreen.left,pGData->rcScreen.right,
		"Y",pGData->rcScreen.top,pGData->rcScreen.bottom,
		pt);
}

/*
	Reads size using ParseWidth and ParseHeight from
	key or key+"Width" and key+"Height"
*/
static void GetRCSize(const char *key,int keylen,SIZE *size,s_GlobalData *pGData)
{
	GetRCPointS(key,keylen,
		"Width",pGData->rcScreen.left,pGData->rcScreen.right,
		"Height",pGData->rcScreen.top,pGData->rcScreen.bottom,
		(POINT *)size);
}

// same as GetRCPoint, but treat as plain numbers rather than screen coords
static void GetRCPointInt(const char *key,int keylen,POINT *pt)
{
	GetRCPointS(key,keylen,
		"X", 0, 0,
		"Y", 0, 0,
		pt);
}

/*
	Accepts:
		standard ls colors
		:12 = GetSysColor(12)
		color*scale
	returns true and sets *dest if valid color is found

	You probably should not include any extra spaces if
	you're scaling a standard ls color, it may get confused.
*/
static bool NewGetRCColor(const char *key,COLORREF *dest)
{
	COLORREF c;
	char *ptr;

	ASSERTV(key!=NULL);
	ASSERTV(dest!=NULL);
	ASSERTV(g_pParseData!=NULL);

	if (GetRCLine(key,g_pParseData->line,MAX_LINE_LENGTH,"")) {

		if (*g_pParseData->line == ':') {
			// system color
			c = GetSysColor(atoi(g_pParseData->line+1));
		} else {
			// standard rc color
			c = GetRCColor(key,-1);
			if (c==-1) return false; // i don't think this will ever happen
		}

		ptr = strchr(g_pParseData->line,'*');
		if (ptr!=NULL) {
			// scale color
			c = ColorScale(c,atoi(ptr+1));
		}

		*dest = c;

		return true;
	} else {
		return false;
	}
}

static bool GetRCBevelColors(const char *key,int keylen,s_BevelColors *dest)
{
	int bevel;

	ASSERTV(key!=NULL);
	ASSERTV(dest!=NULL);
	ASSERTV(g_pParseData!=NULL);
	// Alpha	= xxxAlpha
	// Border = xxxBorderColor
	// Dark		= xxxDarkColor
	// Flat		= xxxColor
	// Light	= xxxLightColor
	//				= xxxBevel
	lstrcpy(g_pParseData->keybuf,key);
	lstrcpy(g_pParseData->keybuf+keylen,"Alpha");
	dest->Alpha = GetRCInt(g_pParseData->keybuf,255);
	lstrcpy(g_pParseData->keybuf+keylen,"Border");
	lstrcpy(g_pParseData->keybuf+keylen+6,"Color");
	NewGetRCColor(g_pParseData->keybuf,&dest->Border);
	lstrcpy(g_pParseData->keybuf+keylen,"Color");
	NewGetRCColor(g_pParseData->keybuf,&dest->Flat);
	lstrcpy(g_pParseData->keybuf+keylen,"Bevel");
	bevel = GetRCInt(g_pParseData->keybuf,64);
	CalcBevelColors(dest,bevel);
	lstrcpy(g_pParseData->keybuf+keylen,"Dark");
	lstrcpy(g_pParseData->keybuf+keylen+4,"Color");
	NewGetRCColor(g_pParseData->keybuf,&dest->Dark);
	lstrcpy(g_pParseData->keybuf+keylen,"Light");
	lstrcpy(g_pParseData->keybuf+keylen+5,"Color");
	NewGetRCColor(g_pParseData->keybuf,&dest->Light);

	return true;
}

// load bitmap and get size
HBITMAP GetRCBitmap(const char *key,SIZE *size)
{
	HBITMAP hbm;
	int w,h;

	ASSERTV(key!=NULL);
	ASSERTV(size!=NULL);

	GetRCString(key,g_pParseData->line,".none",MAX_LINE_LENGTH);
	hbm = LoadLSImage(g_pParseData->line,NULL);
	StatGDIAdd(Bitmap);
	GetLSBitmapSize(hbm,&w,&h);
	size->cx = w;
	size->cy = h;

	return hbm;
}

// read a border definition
void GetRCBorder(const char *key,RECT *dest,const char *def)
{
	int args;

	ASSERTV(key!=NULL);
	ASSERTV(dest!=NULL);
	ASSERTV(def!=NULL);

	GetRCLine(key, g_pParseData->line, MAX_LINE_LENGTH, def);
	args = CommandTokenize(g_pParseData->line,g_pParseData->tokens,4,0);
	if (args == 1) {
		dest->left		=
		dest->top			=
		dest->right		=
		dest->bottom	= atoi(g_pParseData->one);
	} else {
		dest->left		= atoi(g_pParseData->one);
		dest->top			= atoi(g_pParseData->two);
		dest->right		= atoi(g_pParseData->three);
		dest->bottom	= atoi(g_pParseData->four);
	}
}

void GetRCEdgeMap(const char *key, int keylen,s_EdgeMap *em)
{
	ASSERTV(key!=NULL);
	ASSERTV(em!=NULL);

	lstrcpy(g_pParseData->keybuf,key);
	lstrcpy(g_pParseData->keybuf+keylen,"Bmp");
	em->Bitmap = GetRCBitmap(g_pParseData->keybuf,&em->Size);
	lstrcpy(g_pParseData->keybuf+keylen,"Border");
	GetRCBorder(g_pParseData->keybuf,&em->Edge,"0");
	lstrcpy(g_pParseData->keybuf+keylen,"Tile");
	em->Tile = FALSE!=GetRCBool(g_pParseData->keybuf,TRUE);
}

// table of boolean values to read
rcbooldef RCBools[] = {
	// DEFBOOL(n) =
	//  VWMn in step.rc
	//  GlobalData.bn in code
	// true/false is whether or not the setting can be
	// changed and updated with just a repaint. Settings
	// that need more than this can't be used with !VWMToggle
	DEFBOOL(AlwaysOnTop,false)
	DEFBOOL(BackBevel,true)
	DEFBOOL(Bevel,true)
	DEFBOOL(ConstantUpdate,true)
	DEFBOOL(DetectTitleChange,true)
//	DEFBOOL(FixedSizeWindows,true)
//	DEFBOOL(FocusCenter,true)
	DEFBOOL(FocusLast,true)
	DEFBOOL(FocusTop,true)
	DEFBOOL(ForceIconBorderTop,true)
	DEFBOOL(HideTaskOnSwitch,false)
  DEFBOOL(KeepWindowsWithParent,true)
	DEFBOOL(NoBorder,true)
	DEFBOOL(NoDrawSelBack,true)
	DEFBOOL(NoGathering,true)
	DEFBOOL(NoShellFlash,true)
	DEFBOOL(NoShow,false)
	DEFBOOL(NoSwitchOnFocus,true)
	DEFBOOL(NoWindows,true)
	DEFBOOL(NoWrap,true)
	DEFBOOL(ReturnToFirstScreen,true)
	DEFBOOL(ScaleIconToFit,true)
	DEFBOOL(SelBevel,true)
	DEFBOOL(ShowIcons,true)
	DEFBOOL(SnapWindowOnDrag,true)
	DEFBOOL(SwitchOnAnyFocus,true)
	DEFBOOL(Titlebars,true)
	DEFBOOL(TitleForTitle,true)
	DEFBOOL(Transparent,false)
	DEFBOOL(UseFakeTrans,true)
	DEFBOOL(UseTransBlt,true)
	{ 0,0 }
};

// table of actions to read
rcactiondef RCActions[] = {
	// DEFACTION(n) =
	//  VWMn in step.rc
	//  GlobalData.n in code
	DEFACTION(ClickLeft)
	DEFACTION(ClickRight)
	DEFACTION(ClickMiddle)
	// how do i tell which button did the stuff?
	// note to self: finish this now that I have IDropTarget
#ifdef INCLUDE_DROP_EVENTS
	DEFACTION(DropLeft)
	DEFACTION(DropRight)
	DEFACTION(DropMiddle)
#endif
	DEFACTION(MouseOver)
	DEFACTION(MouseOut)
  DEFACTION(MouseMove)
	DEFACTION(Hover)
	DEFACTION(Unhover)
	DEFACTION(Init)
	DEFACTION(Quit)
	DEFACTION(NewWindow)
	DEFACTION(RemoveWindow)
	DEFACTION(SwitchFrom)
	DEFACTION(SwitchTo)
	DEFACTION(TitleChange)
	DEFACTION(MoveWindow)
	DEFACTION(SizeWindow)
#ifdef INCLUDE_SCROLL_EVENTS
	DEFACTION(ScrollUp)
	DEFACTION(ScrollDown)
#endif
	DEFACTION(Flash)
  DEFACTION(Repaint)
	{ 0,0 }
};

//
int ReadConfig(void)
{
	POINT pt;
	s_ParseData p;
	FILE *cfile=NULL;

	GlobalData.Painter = NULL;

	g_pParseData = &p; /* allocated on the stack rather than new/delete */ 
	g_pParseData->tokens[0] = g_pParseData->one;
	g_pParseData->tokens[1] = g_pParseData->two;
	g_pParseData->tokens[2] = g_pParseData->three;
	g_pParseData->tokens[3] = g_pParseData->four;
	g_pParseData->tokens[4] = g_pParseData->five;

	if (GlobalData.ConfigFile) {
		/* check if it's a valid path */ 
		cfile = fopen(GlobalData.ConfigFile,"r");
		if (cfile) {
			fclose(cfile);
			cfile = LCOpen(GlobalData.ConfigFile);
		} else {
			TRACE(TRACE_WARNING,"Rejecting config file.");
			sprintf(g_pParseData->line,"Can't open config file, check your path:\n%s",GlobalData.ConfigFile);
			MessageBox(GlobalData.hParentWnd,
				g_pParseData->line,
				szAppName,MB_OK|MB_ICONERROR);
			/* don't try it again */ 
			delete[] (char *)GlobalData.ConfigFile;			
			GlobalData.ConfigFile = NULL; /* used as a flag */ 
		}
	}

	if (GetRCString("VWMConfigFile",g_pParseData->line,"",MAX_LINE_LENGTH))
	{	/* load config from an alternate file */ 
		if (GlobalData.ConfigFile) delete[] (char *)GlobalData.ConfigFile;
		GlobalData.ConfigFile = new char[lstrlen(g_pParseData->line)+1];
		lstrcpy((char *)GlobalData.ConfigFile,g_pParseData->line);
		if (cfile) LCClose(cfile);
		cfile = LCOpen(GlobalData.ConfigFile);
	}

#ifdef RABIDVWM_MULTIMON
	GetRCString("VWMOnMonitor",g_pParseData->line,"",MAX_LINE_LENGTH);
#if !defined(NDEBUG) && defined(TRACE_FILE)
	sprintf(g_pParseData->one,"on monitor '%s'",g_pParseData->line);
	TRACE(TRACE_STATUS,g_pParseData->one);
#endif
	if (!LSGetMonitorScreenRect(LSGetMonitor(g_pParseData->line),&GlobalData.rcScreen))
	{
		TRACE(TRACE_STATUS,"Using fake monitor.");
		GlobalData.rcScreen.left = -GlobalData.ScreenOffset.x;
		GlobalData.rcScreen.top = -GlobalData.ScreenOffset.y;
		GlobalData.rcScreen.right = GlobalData.rcScreen.left+GlobalData.ScreenSize.cx;
		GlobalData.rcScreen.bottom = GlobalData.rcScreen.top+GlobalData.ScreenSize.cy;
	}

	GetRCString("VWMSwitchMonitor",g_pParseData->line,"all",MAX_LINE_LENGTH);
	GlobalData.hMonitorToSwitch = LSGetMonitorHandle(LSGetMonitor(g_pParseData->line));
#if !defined(NDEBUG) && defined(TRACE_FILE)
	sprintf(g_pParseData->one,"Switching monitor %08X, '%s'",GlobalData.hMonitorToSwitch,g_pParseData->line);
	TRACE(TRACE_STATUS,g_pParseData->one);
#endif
#else /* if not multimonitor */ 
	GlobalData.rcScreen.left = -GlobalData.ScreenOffset.x;
	GlobalData.rcScreen.top = -GlobalData.ScreenOffset.y;
	GlobalData.rcScreen.right = GlobalData.rcScreen.left+GlobalData.ScreenSize.cx;
	GlobalData.rcScreen.bottom = GlobalData.rcScreen.top+GlobalData.ScreenSize.cy;
#endif /* RABIDVWM_MULTIMON */ 

	pt.x = 16; pt.y = 12;
	GetRCPointInt(  "VWMDeskSpacing",14,&pt);
	GlobalData.DeskSize.cx = GlobalData.ScreenSize.cx+pt.x;
	GlobalData.DeskSize.cy = GlobalData.ScreenSize.cy+pt.y;
	GlobalData.DeskOffset.x = GlobalData.ScreenOffset.x-pt.x/2;
	GlobalData.DeskOffset.y = GlobalData.ScreenOffset.y-pt.y/2;

	/* read bools */ 
	{
		rcbooldef *pdef;
		lstrcpy(g_pParseData->keybuf,RCPREFIX);
		pdef = RCBools;
		while (pdef->key != NULL) {
			lstrcpy(g_pParseData->keybuf+RCPREFIXLEN,pdef->key);
			GlobalData.*pdef->pItem = FALSE!=GetRCBool(g_pParseData->keybuf,TRUE);
			pdef++;
		}
	}
#if !defined(NDEBUG) && defined(TRACE_FILE)
	DumpBools();
#endif

	GlobalData.VWMPos.x = GlobalData.VWMPos.y = 0;
	GlobalData.VWMSize.cx = GlobalData.VWMSize.cy = 0x7fffffff;
	/* using key length of 3 to get VWMX & VWMY instead of VWMPosX and VWMPosY */ 
	GetRCPoint(			"VWMPos",			3,&GlobalData.VWMPos,&GlobalData);
	GetRCSize(			"VWMSize",		3,&GlobalData.VWMSize,&GlobalData);

	/* default size                                                            */ 
	if (GlobalData.VWMSize.cx==0x7fffffff || GlobalData.VWMSize.cy==0x7fffffff)
	{
		/* set to size of parent, if any                                         */ 
		GlobalData.VWMSize.cx = GlobalData.ParentRect.right-GlobalData.ParentRect.left;
		GlobalData.VWMSize.cy = GlobalData.ParentRect.bottom-GlobalData.ParentRect.top;
		TRACE(TRACE_STATUS,"Size not specified, filling parent.");
		if (!GlobalData.bStayChildWindow || /* don't use parent size if it's the desktop */ 
			GlobalData.VWMSize.cx==0 && GlobalData.VWMSize.cy==0) {
			/* unparented, user really should set this, but hasn't                 */ 
			/* use an arbitrary value to make sure we don't just fail              */ 
			GlobalData.VWMSize.cx =
			GlobalData.VWMSize.cy = 100;
			TRACE(TRACE_WARNING,"No parent, using 100x100.");
			/* if we start hidden, maybe they'll resize before showing             */ 
			if (!GlobalData.bNoShow)
				MessageBox(NULL,
					"Please specify VWMSize or VWMWidth and VWMHeight\nwhen not loading in a box or wharf.",
					szAppName,MB_OK);
		} else {
			/* default to not overlapping the edge                                 */ 
			GlobalData.VWMSize.cx -= GlobalData.VWMPos.x;
			GlobalData.VWMSize.cy -= GlobalData.VWMPos.y;
		}
	}

#if !defined(NDEBUG) && defined(TRACE_FILE)
	{
		sprintf(g_pParseData->line,"parent = { %5d, %5d, %5d, %5d }",
			GlobalData.ParentRect.left,GlobalData.ParentRect.top,
			GlobalData.ParentRect.right,GlobalData.ParentRect.bottom);
		TRACE(TRACE_STATUS,g_pParseData->line);
		sprintf(g_pParseData->line,"screen = { %5d, %5d, %5d, %5d }",
			GlobalData.rcScreen.left,GlobalData.rcScreen.top,
			GlobalData.rcScreen.right,GlobalData.rcScreen.bottom);
		TRACE(TRACE_STATUS,g_pParseData->line);
		sprintf(g_pParseData->line,"vwm =  <%5d, %5d> x <%5d, %5d>",
			GlobalData.VWMPos.x,GlobalData.VWMPos.y,
			GlobalData.VWMSize.cx, GlobalData.VWMSize.cy);
		TRACE(TRACE_STATUS,g_pParseData->line);
	}
#endif

	pt.x = pt.y = 0;
	GetRCPointInt(	"VWMDesks",		8,&pt);
/*	GlobalData.PerDeskSize.cx =
	GlobalData.PerDeskSize.cy = 0;
//	GetRCPointInt(	"VWMDeskSize",11,&GlobalData.PerDeskSize);
*/
	if (pt.x==0 && pt.y==0/*
		&& (GlobalData.PerDeskSize.cx==0
			|| GlobalData.PerDeskSize.cy==0)*/) {
		MessageBox(NULL,
			"Number of desks specified was 0.\n"
			"Make sure you've set VWMDesks.", // or VWMDeskSize.", ** not done yet
			szAppName,MB_OK);
		GlobalData.bGuessDeskCount = true;
	}

	if (pt.x<=0) pt.x=1;
	if (pt.y<=0) pt.y=1;

	GlobalData.nDesksX = pt.x;
	GlobalData.nDesksY = pt.y;

	/* actual flag is irrelevant because SetOnTop ignores it in this case */ 
	CTRACE(TRACE_WARNING,GlobalData.bLoadInBox,"Always on top while in box.");
	if (!GlobalData.bNoShow)
		GlobalData.bNoShow = FALSE!=GetRCBool("VWMHidden",TRUE);

	/* read actions */ 
	{
		rcactiondef *pdef;
		/* prefix is already there from bool loop  */ 
		/* lstrcpy(g_pParseData->keybuf,RCPREFIX); */ 
		pdef = RCActions;
    lstrcpy(g_pParseData->keybuf+RCPREFIXLEN,"On");
		while (pdef->key != NULL) {
			lstrcpy(g_pParseData->keybuf+RCPREFIXLEN+2,pdef->key);
			/* if the action has been set              */ 
			if (GetRCLine(g_pParseData->keybuf,g_pParseData->line,MAX_LINE_LENGTH,NULL)) {
				/* create the action                     */ 
				GlobalData.*pdef->pAction = new s_Action;
				ZeroMemory(GlobalData.*pdef->pAction,sizeof(s_Action));
				/* allocate space for the command string */ 
				(GlobalData.*pdef->pAction)->Command = new char[lstrlen(g_pParseData->line)+1];
				/* copy the command string               */ 
				/* (cast from const, ok to set here)     */ 
				lstrcpy(const_cast<char *>((GlobalData.*pdef->pAction)->Command),g_pParseData->line);
			}
			pdef++;
		}
	}
#if !defined(NDEBUG) && defined(TRACE_FILE)
	DumpActions();
#endif

	/* read colors */ 

	/* read autobevel values */ 

	/* read old inflexible version for backwards compatibility */ 
	GlobalData.cWnd.Flat = GetSysColor(COLOR_3DFACE);
	NewGetRCColor("VWMForeColor",&GlobalData.cWnd.Flat);
	if (GlobalData.bBevel) /* old behavior is strange */ 
		NewGetRCColor("VWMWinColor",&GlobalData.cWnd.Flat);
	GlobalData.cBack.Border = GetSysColor(COLOR_WINDOWFRAME);
	NewGetRCColor("VWMBorderColor",&GlobalData.cBack.Border);
	GlobalData.cTitle.Flat =
	GlobalData.cTitle.Border =
	GlobalData.cWnd.Border = GlobalData.cWnd.Border;
	if (GlobalData.bNoBorder) { /* turn off borders */ 
		/* actually, just draw them the same color, it's easier */ 
		GlobalData.cBack.Border = GlobalData.cWnd.Flat;
		GlobalData.cWnd.Border = GlobalData.cBack.Flat;
		/* title is already the same                            */ 
	}
	GlobalData.cBack.Flat = GetSysColor(COLOR_3DFACE);
	NewGetRCColor("VWMBackColor",&GlobalData.cBack.Flat);
	GlobalData.cSelect.Flat = GetSysColor(COLOR_HIGHLIGHT);
	NewGetRCColor("VWMSelBackColor",&GlobalData.cSelect.Flat);
	GlobalData.cFlash.Flat = GlobalData.cSelect.Flat;
	NewGetRCColor("VWMFlashColor",&GlobalData.cFlash.Flat);

	/* new colors */ 
	if (NewGetRCColor("VWMTitleColor",&GlobalData.cTitle.Flat)) {
		if (GlobalData.bNoBorder)
			GlobalData.cTitle.Border = GlobalData.cTitle.Flat;
	}

	GetRCBevelColors("VWMBack",4+3,&GlobalData.cBack);
	GetRCBevelColors("VWMSel",3+3,&GlobalData.cSelect);
	GetRCBevelColors("VWMWin",3+3,&GlobalData.cWnd);
	GetRCBevelColors("VWMTitle",5+3,&GlobalData.cTitle);
	GetRCBevelColors("VWMFlash",5+3,&GlobalData.cFlash);
	GlobalData.DoubleBevelWidth = GetRCInt("VWMBevelWidth",0);

	/* read mouse actions */ 
	GlobalData.MouseAction[0] = GetRCInt("VWMMouseLeft",2);
	GlobalData.MouseAction[1] = GetRCInt("VWMMouseRight",1);
	GlobalData.MouseAction[2] = GetRCInt("VWMMouseMiddle",0);
	if (GlobalData.MouseAction[0] > MOUSE_ACTION_COUNT) GlobalData.MouseAction[0]=0;
	if (GlobalData.MouseAction[1] > MOUSE_ACTION_COUNT) GlobalData.MouseAction[1]=0;
	if (GlobalData.MouseAction[2] > MOUSE_ACTION_COUNT) GlobalData.MouseAction[2]=0;

	/* read titlebar height via mod or pixel value */ 
	GlobalData.TitlebarHeight = GetRCInt("VWMTitlebarMod",0);
	if (GlobalData.TitlebarHeight!=0)
		GlobalData.TitlebarHeight = GlobalData.DeskSize.cy / GlobalData.TitlebarHeight;
	else
		GlobalData.TitlebarHeight = 4;
	GlobalData.TitlebarHeight = GetRCInt("VWMTitlebarHeight",GlobalData.TitlebarHeight);

	/* read icon settings */ 
	GlobalData.IconSize = GetRCInt("VWMIconSize",16);
	GlobalData.MinIconSize = GetRCInt("VWMMinIconSize",4);
	GetRCBorder("VWMIconBorder",&GlobalData.IconBorder,"2");

	/* read bitmaps */ 
	GetRCEdgeMap("VWMBack",7,&GlobalData.emBack);
	GetRCBorder("VWMDeskBorder",&GlobalData.rDeskBorder,"0");
	GetRCEdgeMap("VWMSel",6,&GlobalData.emSel);
	GetRCBorder("VWMSelExpand",&GlobalData.rSelExpand,"0");

	GetRCEdgeMap("VWMWin",6,&GlobalData.emWnd);
	/* always load images in case the option is !VWMToggled */ 
	GetRCEdgeMap("VWMTitlebar",11,&GlobalData.emTitle);
	GetRCEdgeMap("VWMFlash",8,&GlobalData.emFlash);
	GetRCEdgeMap("VWMFlashTitle",13,&GlobalData.emFlashTitle);

	if (!GlobalData.emWnd.Bitmap)
	{ /* if the bitmap didn't load, maybe it's .snapshot? */ 
		GetRCString("VWMWinBmp", g_pParseData->line, "", MAX_LINE_LENGTH);
		if (*g_pParseData->line=='.') {
			/* check snap shot level */ 
			*g_pParseData->two = 0;
			CommandTokenize(g_pParseData->line,g_pParseData->tokens,2,NULL);
			if (lstrcmp(g_pParseData->one,".snapshot")==0)
			{
				GlobalData.Painter = &Paint_SnapShot;
				GlobalData.bTitlebars = false;
				GlobalData.SnapShotLevel = atoi(g_pParseData->two)+1;
			}
		}
	}
	if (!GlobalData.emBack.Bitmap)
	{ /* if the bitmap didn't load, maybe it's .snapshot? */ 
		GetRCString("VWMBackBmp", g_pParseData->line, "", MAX_LINE_LENGTH);
		if (*g_pParseData->line=='.') {
			/* check snap shot level */ 
			*g_pParseData->two = 0;
			CommandTokenize(g_pParseData->line,g_pParseData->tokens,2,NULL);
			if (lstrcmp(g_pParseData->one,".snapshot")==0)
			{
				GlobalData.PaintVWM = &PaintVWM_SnapShot;
				GlobalData.SnapShotLevelVWM = atoi(g_pParseData->two)+1;
			}
		}
	}
	if (!GlobalData.emSel.Bitmap)
	{ /* if the bitmap didn't load, maybe it's .snapshot? */ 
		GetRCString("VWMSelBmp", g_pParseData->line, "", MAX_LINE_LENGTH);
		if (*g_pParseData->line=='.') {
			/* check snap shot level */ 
			*g_pParseData->two = 0;
			CommandTokenize(g_pParseData->line,g_pParseData->tokens,2,NULL);
			if (lstrcmp(g_pParseData->one,".snapshot")==0)
			{
				GlobalData.PaintSelect = &PaintSelect_SnapShot;
				GlobalData.SnapShotLevelSel = atoi(g_pParseData->two)+1;
			}
		}
	}

  GlobalData.InitialDesk = DESK_NONE;
	if (GetRCLine("VWMInitialDesk", g_pParseData->line, MAX_LINE_LENGTH, "")) {
		/* this will accept windowspecs, but   */ 
		/* they'll only work on !refresh       */ 
		/* or if an lsbox is destroyed/created */ 
		GlobalData.InitialDesk = GetBangDeskNo(g_pParseData->line,NULL);
	}

	GlobalData.HoverTime = GetRCInt("VWMHoverTime",500);
	GlobalData.RefreshTime = GetRCInt("VWMRefreshTime",250);

	/* default to flash forever */ 
	GlobalData.iFlashCount = GetRCInt("VWMFlashCount",0);

	/* select miniwindow painter */ 
	if (GlobalData.Painter == NULL) {
		if (GlobalData.emWnd.Bitmap &&
			(!GlobalData.bTitlebars || GlobalData.emTitle.Bitmap))
			GlobalData.Painter = &Paint_Bitmap;
		else
			GlobalData.Painter = &Paint_Colors;
	}

	if (GlobalData.PaintVWM==NULL) {
		/* select vwm window painter */ 
		if (GlobalData.emBack.Bitmap)
			GlobalData.PaintVWM = &PaintVWM_Bitmap;
		else
			GlobalData.PaintVWM = &PaintVWM_Colors;
	}

	if (GlobalData.PaintSelect==NULL) {
		if (GlobalData.emSel.Bitmap)
			GlobalData.PaintSelect = &PaintSelect_Bitmap;
		else
			GlobalData.PaintSelect = &PaintSelect_Colors;
	}

#if !defined(NDEBUG) && defined(TRACE_FILE)
	{
		char b[256];

		if (GlobalData.Painter==&Paint_Bitmap)
			TRACE(TRACE_DUMP,    "MiniWnd Painter = BITMAP");
		else if (GlobalData.Painter==&Paint_Colors)
			TRACE(TRACE_DUMP,    "MiniWnd Painter = COLORS");
		else if (GlobalData.Painter==&Paint_SnapShot) {
			sprintf(b,"MiniWnd Painter = SNAPSHOT %i",GlobalData.SnapShotLevel); TRACE(TRACE_DUMP,b); }
		else {
			sprintf(b,"MiniWnd Painter = %08X",GlobalData.Painter); TRACE(TRACE_DUMP,b); }

		if (GlobalData.PaintVWM==&PaintVWM_Bitmap)
			TRACE(TRACE_DUMP,    "    VWM Painter = BITMAP");
		else if (GlobalData.PaintVWM==&PaintVWM_Colors)
			TRACE(TRACE_DUMP,    "    VWM Painter = COLORS");
		else if (GlobalData.PaintVWM==&PaintVWM_SnapShot) {
			sprintf(b,"MiniWnd Painter = SNAPSHOT %i",GlobalData.SnapShotLevelVWM); TRACE(TRACE_DUMP,b); }
		else {
			sprintf(b,"    VWM Painter = %08X",GlobalData.PaintVWM); TRACE(TRACE_DUMP,b); }

		if (GlobalData.PaintSelect==&PaintSelect_Bitmap)
			TRACE(TRACE_DUMP,    " Select Painter = BITMAP");
		else if (GlobalData.PaintSelect==&PaintSelect_Colors)
			TRACE(TRACE_DUMP,    " Select Painter = COLORS");
		else if (GlobalData.PaintSelect==&PaintSelect_SnapShot) {
			sprintf(b,"MiniWnd Painter = SNAPSHOT %i",GlobalData.SnapShotLevelSel); TRACE(TRACE_DUMP,b); }
		else {
			sprintf(b," Select Painter = %08X",GlobalData.PaintSelect); TRACE(TRACE_DUMP,b); }
	}
#endif

	if (cfile)
		LCClose(cfile);

	ReadStarLists();

  if (!dUpdateLayeredWindow) GlobalData.bTransparent = false;
  if (!dAlphaBlend)
  {
    GlobalData.bUseFakeTrans = false;
    GlobalData.cBack.Alpha = 255;
    GlobalData.cFlash.Alpha = 255;
    GlobalData.cSelect.Alpha = 255;
    GlobalData.cTitle.Alpha = 255;
    GlobalData.cWnd.Alpha = 255;
  }

	return 0;
}

// clean up things created in ReadConfig
int ReleaseConfig(void)
{
	DI;

	UnreadStarLists();

	/* clean up actions */ 
	{
		TRACE(TRACE_STATUS,"Removing actions");
		rcactiondef *pdef;
		pdef = RCActions;
		while (pdef->key != NULL) {
			if (GlobalData.*pdef->pAction) {
				TRACE(TRACE_STATUS,(GlobalData.*pdef->pAction)->Command);
				delete[] (char *)((GlobalData.*pdef->pAction)->Command);
				delete (GlobalData.*pdef->pAction);
				(GlobalData.*pdef->pAction) = NULL; /* used as flag */ 
			}
			pdef++;
		}
	}

	TRACE(TRACE_STATUS,"Deleting GUI objects");
	if (GlobalData.emBack.Bitmap) { CI DeleteObject(GlobalData.emBack.Bitmap); StatGDIDel(Bitmap); }
	if (GlobalData.emSel.Bitmap) { CI DeleteObject(GlobalData.emSel.Bitmap); StatGDIDel(Bitmap); }
	if (GlobalData.emFlashTitle.Bitmap) { CI DeleteObject(GlobalData.emFlashTitle.Bitmap); StatGDIDel(Bitmap); }
	if (GlobalData.emFlash.Bitmap) { CI DeleteObject(GlobalData.emFlash.Bitmap); StatGDIDel(Bitmap); }
	if (GlobalData.emWnd.Bitmap) { CI DeleteObject(GlobalData.emWnd.Bitmap); StatGDIDel(Bitmap); }
	if (GlobalData.emTitle.Bitmap) { CI DeleteObject(GlobalData.emTitle.Bitmap); StatGDIDel(Bitmap); }
	if (GlobalData.Desks) delete[] GlobalData.Desks;
	CTRACE(TRACE_WARNING,!i,"ReleaseConfig: Failed.");
#ifndef NDEBUG
	GlobalData.emBack.Bitmap = NULL;
	GlobalData.emSel.Bitmap = NULL;
	GlobalData.emWnd.Bitmap = NULL;
	GlobalData.emFlash.Bitmap = NULL;
	GlobalData.emFlashTitle.Bitmap = NULL;
	GlobalData.emTitle.Bitmap = NULL;
	GlobalData.Desks = NULL;
#endif

	GlobalData.PaintVWM = NULL;    /* used as flag */ 
	GlobalData.Painter = NULL;     /* used as flag */ 
	GlobalData.PaintSelect = NULL; /* used as flag */ 

	return 0;
}

